import ast
import logging
import pandas as pd
from wiliot_deployment_tools.interface.uart_if import UARTInterface
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.if_defines import *
import time
import datetime
import threading
import binascii

class SnifferPkt():
    def __init__(self, raw_output, time_received, rx_channel):
        self.adva = raw_output[:12]
        self.packet = raw_output[12:74]
        try:
            self.rssi = int.from_bytes(binascii.unhexlify(raw_output[74:]), 'big') * -1
        except ValueError:
            self.rssi = 99
        self.time_received = time_received
        self.rx_channel = rx_channel

    def __repr__(self):
        return f'CH{self.rx_channel}|{self.adva}{self.packet} RSSI:{self.rssi} {self.time_received}'

    def to_dict(self):
        return {'adva': self.adva, 'packet': self.packet, 'rssi': self.rssi, 'time_received': self.time_received, 'rx_channel': self.rx_channel}

class SnifferPkts():
    def __init__(self, pkts=[]):
        self.pkts = pkts

    def __add__(self, other):
        return SnifferPkts(self.pkts + other.pkts)

    def __len__(self):
        return len(self.pkts)

    def __repr__(self):
        return self.pkts

    def process_pkt(self, raw_output, time_received, rx_channel, print_pkt=False, adva_filter=None):
        pkt = SnifferPkt(raw_output, time_received, rx_channel)
        if adva_filter:
            if pkt.adva != adva_filter:
                return None
        self.pkts.append(pkt)
        if print_pkt:
            print(pkt)
        return pkt

    def filter_pkts(self, raw_packet=None, adva=None, time_range:tuple=None):
        result = []
        for pkt in self.pkts:
            if (raw_packet is not None) and (pkt.packet == raw_packet) or \
                (adva is not None) and (pkt.adva == adva) or \
                (time_range is not None) and (time_range[0] < pkt.time_received < time_range[1]):
                result.append(pkt)
        return SnifferPkts(result)

    def flush_pkts(self):
        self.pkts = []

    def to_list(self):
        return [p.to_dict() for p in self.pkts]

    def to_pandas(self):
        return pd.DataFrame().from_dict(self.to_list())

def cur_time():
    return datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')


class BLESniffer():
    def __init__(self, uart:UARTInterface, print_pkt=False, logger_filepath=None, adva_filter=None):
        self.uart = uart
        self.listener_thread = None
        self.sniffer_pkts = SnifferPkts()
        self.listening = False
        self.listener_lock = threading.Lock()
        self.print = print_pkt
        self.rx_channel = 0
        self.adva_filter = None
        if adva_filter is not None:
            adva_filter = adva_filter.upper()
            assert isinstance(adva_filter, str)\
                and len(adva_filter) == 12\
                and all(c in '0123456789ABCDEF' for c in adva_filter), "Input must be a 12-character hexadecimal string"
            self.adva_filter = adva_filter
        # Configure Logger
        logger = logging.getLogger('ble_sniffer')
        logger.setLevel(logging.DEBUG)
        if logger_filepath is not None:
            # create file handler
            fh = logging.FileHandler(logger_filepath)
            fh.setLevel(logging.DEBUG)
            formatter = logging.Formatter('%(message)s')
            fh.setFormatter(formatter)
            logger.addHandler(fh)
            logger.propagate = False # Do not send logs to 'root' logger
            debug_print(f'BLE Sniffer Logger initialized at {logger_filepath}')
        self.logger = logger

    def packet_listener(self):
        while self.listening:
            line = self.uart.read_line()
            if line is not None and len(line) == 76:
                with self.listener_lock:
                    pkt = self.sniffer_pkts.process_pkt(line, datetime.datetime.now(), self.rx_channel, self.print, self.adva_filter)
                    if pkt:
                        self.logger.info(pkt)

    # Change sniffing modes
    def start_sniffer(self, rx_channel):
        self.logger.debug(f'{cur_time()} | Starting Sniffer on CH{rx_channel}')
        self.uart.set_sniffer(rx_channel)
        self.rx_channel = rx_channel
        self.listener_thread = threading.Thread(target=self.packet_listener)
        self.listening = True
        self.listener_thread.start()

    def stop_sniffer(self):
        self.logger.debug(f'{cur_time()} | Stopping Sniffer')
        self.flush_pkts()
        self.listening = False
        if self.listener_thread is not None:
            self.listener_thread.join()
            self.listener_thread = None
            self.uart.cancel_sniffer()
        self.rx_channel = 0

    def reset_sniffer(self, rx_channel):
        self.logger.debug(f'{cur_time()} | Reseting Sniffer')
        self.stop_sniffer()
        self.start_sniffer(rx_channel)
        self.flush_pkts()

    # Data Handling
    def get_all_pkts(self):
        return self.sniffer_pkts

    def get_filtered_packets(self, raw_packet=None, adva=None, time_range:tuple=None):
        return self.sniffer_pkts.filter_pkts(raw_packet, adva, time_range)

    def flush_pkts(self):
        self.logger.debug(f'{cur_time()} | Flushing packets')
        self.sniffer_pkts.flush_pkts()

    def to_pandas(self):
        return self.sniffer_pkts.to_pandas()

    # Packet Counters
    def get_pkts_cntrs(self, channel, lookout_time=DEFAULT_LOOKOUT_TIME):
        self.logger.debug(f'{cur_time()} | Getting pkt counters for CH{channel}')
        if self.rx_channel != channel:
            self.uart.set_rx(channel)
            self.rx_channel = channel
        
        lookout_time = lookout_time
        pkt_cntrs = None
        self.uart.flush()
        self.uart.write_ble_command(GET_LOGGER_COUNTERS)
        start_time = time.time()
        current_time = start_time
        while current_time - start_time < lookout_time:
            line = self.uart.read_line()
            if line is not None and "'bad_crc'" in line:
                start_of_cntr_index = line.find('{')
                pkt_cntrs = line[start_of_cntr_index:]
                debug_print(f"pkt_cntrs: {pkt_cntrs}")
                return ast.literal_eval(pkt_cntrs)
            current_time = time.time()
        debug_print(f"No counter received within the time limit of {lookout_time} seconds")
        return pkt_cntrs

class BLESnifferContext():
    def __init__(self, ble_sniffer:BLESniffer, rx_channel):
        self.ble_sniffer = ble_sniffer
        self.rx_channel = rx_channel

    def __enter__(self):
        self.ble_sniffer.flush_pkts()
        self.ble_sniffer.start_sniffer(self.rx_channel)
        return self.ble_sniffer

    def __exit__(self, exc_type, exc_value, traceback):
        if self.ble_sniffer:
            self.ble_sniffer.stop_sniffer()


RX_CHANNELS = [37, 38, 39]
RX_CHANNELS_SNIFFING_KIT = list(range(40))
SET_SNIFFER = '!set_logger_mode 2'
CANCEL_SNIFFER = '!set_logger_mode 0'
GET_LOGGER_COUNTERS = '!get_logger_counters'
GATEWAY_APP = '!gateway_app'
STOP_ADVERTISING = '!stop_advertising'
CANCEL = '!cancel'
RESET_GW = '!reset'
VERSION = '!version'
SEND_ALL_ADV_CHANNELS = 0
SERIAL_TIMEOUT = 0.1
SEP = "#"*100
LENGTH = '1E'
GAP_TYPE = '16'
DEFAULT_ADVA = 'FFFFFFFFFFFF'
DEFAULT_DUPLICATES = 3
GW_APP_VERSION_HEADER = 'WILIOT_GW_BLE_CHIP_SW_VER'
LOCATION = 'location'
DEFAULT_OUTPUT_POWER = 8
DEFAULT_DELAY = 20
BRIDGES = [i for i in range(3)]
MAX_NFPKT = 65535
MAX_RSSI = 255
ADVA_LENGTH = 12
GAP_LENGTH = 4
SERVICE_UUID_LENGTH = 4
GROUP_ID_LENGTH = 6
UPLINK_DUPLICATIONS = [i for i in range(3, 7)]
UPLINK_TIME_DELAYS = [20, 130, 255]
UNIFIED_DUPLICATIONS = [i for i in range(3, 7)]
UNIFIED_TIME_DELAYS = [20, 130, 255]
DEFAULT_LOOKOUT_TIME = 2



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Location(_message.Message):
    __slots__ = ("lat", "lng")
    LAT_FIELD_NUMBER: _ClassVar[int]
    LNG_FIELD_NUMBER: _ClassVar[int]
    lat: float
    lng: float
    def __init__(self, lat: _Optional[float] = ..., lng: _Optional[float] = ...) -> None: ...

class Value(_message.Message):
    __slots__ = ("integerValue", "numberValue", "stringValue", "boolValue")
    INTEGERVALUE_FIELD_NUMBER: _ClassVar[int]
    NUMBERVALUE_FIELD_NUMBER: _ClassVar[int]
    STRINGVALUE_FIELD_NUMBER: _ClassVar[int]
    BOOLVALUE_FIELD_NUMBER: _ClassVar[int]
    integerValue: int
    numberValue: float
    stringValue: str
    boolValue: bool
    def __init__(self, integerValue: _Optional[int] = ..., numberValue: _Optional[float] = ..., stringValue: _Optional[str] = ..., boolValue: bool = ...) -> None: ...

class GatewayData(_message.Message):
    __slots__ = ("gatewayId", "timestamp", "packets", "location")
    class Packet(_message.Message):
        __slots__ = ("payload", "timestamp", "sequenceId", "rssi", "aliasBridgeId")
        PAYLOAD_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
        SEQUENCEID_FIELD_NUMBER: _ClassVar[int]
        RSSI_FIELD_NUMBER: _ClassVar[int]
        ALIASBRIDGEID_FIELD_NUMBER: _ClassVar[int]
        payload: bytes
        timestamp: int
        sequenceId: int
        rssi: int
        aliasBridgeId: str
        def __init__(self, payload: _Optional[bytes] = ..., timestamp: _Optional[int] = ..., sequenceId: _Optional[int] = ..., rssi: _Optional[int] = ..., aliasBridgeId: _Optional[str] = ...) -> None: ...
    GATEWAYID_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    PACKETS_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    gatewayId: str
    timestamp: int
    packets: _containers.RepeatedCompositeFieldContainer[GatewayData.Packet]
    location: Location
    def __init__(self, gatewayId: _Optional[str] = ..., timestamp: _Optional[int] = ..., packets: _Optional[_Iterable[_Union[GatewayData.Packet, _Mapping]]] = ..., location: _Optional[_Union[Location, _Mapping]] = ...) -> None: ...

class UplinkMessage(_message.Message):
    __slots__ = ("gatewayStatus", "gatewayInfo", "gatewayLogs", "actionStatus")
    GATEWAYSTATUS_FIELD_NUMBER: _ClassVar[int]
    GATEWAYINFO_FIELD_NUMBER: _ClassVar[int]
    GATEWAYLOGS_FIELD_NUMBER: _ClassVar[int]
    ACTIONSTATUS_FIELD_NUMBER: _ClassVar[int]
    gatewayStatus: GatewayStatus
    gatewayInfo: GatewayInfo
    gatewayLogs: GatewayLogs
    actionStatus: ActionStatus
    def __init__(self, gatewayStatus: _Optional[_Union[GatewayStatus, _Mapping]] = ..., gatewayInfo: _Optional[_Union[GatewayInfo, _Mapping]] = ..., gatewayLogs: _Optional[_Union[GatewayLogs, _Mapping]] = ..., actionStatus: _Optional[_Union[ActionStatus, _Mapping]] = ...) -> None: ...

class GatewayStatus(_message.Message):
    __slots__ = ("gatewayId", "gatewayType", "downlinkSupported", "bridgeOtaUpgradeSupported", "apiVersion", "version", "bleSwVersion", "interfaceSwVersion", "location", "config", "bleAddress")
    class ConfigEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Value
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[Value, _Mapping]] = ...) -> None: ...
    GATEWAYID_FIELD_NUMBER: _ClassVar[int]
    GATEWAYTYPE_FIELD_NUMBER: _ClassVar[int]
    DOWNLINKSUPPORTED_FIELD_NUMBER: _ClassVar[int]
    BRIDGEOTAUPGRADESUPPORTED_FIELD_NUMBER: _ClassVar[int]
    APIVERSION_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    BLESWVERSION_FIELD_NUMBER: _ClassVar[int]
    INTERFACESWVERSION_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    BLEADDRESS_FIELD_NUMBER: _ClassVar[int]
    gatewayId: str
    gatewayType: str
    downlinkSupported: bool
    bridgeOtaUpgradeSupported: bool
    apiVersion: int
    version: str
    bleSwVersion: str
    interfaceSwVersion: str
    location: Location
    config: _containers.MessageMap[str, Value]
    bleAddress: str
    def __init__(self, gatewayId: _Optional[str] = ..., gatewayType: _Optional[str] = ..., downlinkSupported: bool = ..., bridgeOtaUpgradeSupported: bool = ..., apiVersion: _Optional[int] = ..., version: _Optional[str] = ..., bleSwVersion: _Optional[str] = ..., interfaceSwVersion: _Optional[str] = ..., location: _Optional[_Union[Location, _Mapping]] = ..., config: _Optional[_Mapping[str, Value]] = ..., bleAddress: _Optional[str] = ...) -> None: ...

class GatewayInfo(_message.Message):
    __slots__ = ("entries",)
    class EntriesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Value
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[Value, _Mapping]] = ...) -> None: ...
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    entries: _containers.MessageMap[str, Value]
    def __init__(self, entries: _Optional[_Mapping[str, Value]] = ...) -> None: ...

class GatewayLogs(_message.Message):
    __slots__ = ("logs",)
    LOGS_FIELD_NUMBER: _ClassVar[int]
    logs: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, logs: _Optional[_Iterable[str]] = ...) -> None: ...

class ActionStatus(_message.Message):
    __slots__ = ("action", "status")
    ACTION_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    action: int
    status: int
    def __init__(self, action: _Optional[int] = ..., status: _Optional[int] = ...) -> None: ...

class DownlinkMessage(_message.Message):
    __slots__ = ("txPacket", "gatewayAction", "bridgeUpgrade", "gatewayConfig", "customBroker")
    TXPACKET_FIELD_NUMBER: _ClassVar[int]
    GATEWAYACTION_FIELD_NUMBER: _ClassVar[int]
    BRIDGEUPGRADE_FIELD_NUMBER: _ClassVar[int]
    GATEWAYCONFIG_FIELD_NUMBER: _ClassVar[int]
    CUSTOMBROKER_FIELD_NUMBER: _ClassVar[int]
    txPacket: TxPacket
    gatewayAction: GatewayAction
    bridgeUpgrade: BridgeUpgrade
    gatewayConfig: GatewayConfig
    customBroker: CustomBroker
    def __init__(self, txPacket: _Optional[_Union[TxPacket, _Mapping]] = ..., gatewayAction: _Optional[_Union[GatewayAction, _Mapping]] = ..., bridgeUpgrade: _Optional[_Union[BridgeUpgrade, _Mapping]] = ..., gatewayConfig: _Optional[_Union[GatewayConfig, _Mapping]] = ..., customBroker: _Optional[_Union[CustomBroker, _Mapping]] = ...) -> None: ...

class TxPacket(_message.Message):
    __slots__ = ("payload", "maxRetries", "maxDurationMs")
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    MAXRETRIES_FIELD_NUMBER: _ClassVar[int]
    MAXDURATIONMS_FIELD_NUMBER: _ClassVar[int]
    payload: bytes
    maxRetries: int
    maxDurationMs: int
    def __init__(self, payload: _Optional[bytes] = ..., maxRetries: _Optional[int] = ..., maxDurationMs: _Optional[int] = ...) -> None: ...

class GatewayAction(_message.Message):
    __slots__ = ("action",)
    ACTION_FIELD_NUMBER: _ClassVar[int]
    action: str
    def __init__(self, action: _Optional[str] = ...) -> None: ...

class BridgeUpgrade(_message.Message):
    __slots__ = ("rebootPacket", "txMaxDurationMs", "txMaxRetries", "bridgeId", "versionUuid", "upgradeBlSd", "imageDirUrl")
    REBOOTPACKET_FIELD_NUMBER: _ClassVar[int]
    TXMAXDURATIONMS_FIELD_NUMBER: _ClassVar[int]
    TXMAXRETRIES_FIELD_NUMBER: _ClassVar[int]
    BRIDGEID_FIELD_NUMBER: _ClassVar[int]
    VERSIONUUID_FIELD_NUMBER: _ClassVar[int]
    UPGRADEBLSD_FIELD_NUMBER: _ClassVar[int]
    IMAGEDIRURL_FIELD_NUMBER: _ClassVar[int]
    rebootPacket: bytes
    txMaxDurationMs: int
    txMaxRetries: int
    bridgeId: str
    versionUuid: str
    upgradeBlSd: bool
    imageDirUrl: str
    def __init__(self, rebootPacket: _Optional[bytes] = ..., txMaxDurationMs: _Optional[int] = ..., txMaxRetries: _Optional[int] = ..., bridgeId: _Optional[str] = ..., versionUuid: _Optional[str] = ..., upgradeBlSd: bool = ..., imageDirUrl: _Optional[str] = ...) -> None: ...

class GatewayConfig(_message.Message):
    __slots__ = ("version", "bleSwVersion", "interfaceSwVersion", "location", "config")
    class ConfigEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Value
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[Value, _Mapping]] = ...) -> None: ...
    VERSION_FIELD_NUMBER: _ClassVar[int]
    BLESWVERSION_FIELD_NUMBER: _ClassVar[int]
    INTERFACESWVERSION_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    version: str
    bleSwVersion: str
    interfaceSwVersion: str
    location: Location
    config: _containers.MessageMap[str, Value]
    def __init__(self, version: _Optional[str] = ..., bleSwVersion: _Optional[str] = ..., interfaceSwVersion: _Optional[str] = ..., location: _Optional[_Union[Location, _Mapping]] = ..., config: _Optional[_Mapping[str, Value]] = ...) -> None: ...

class CustomBroker(_message.Message):
    __slots__ = ("customBroker", "port", "brokerUrl", "username", "password", "updateTopic", "statusTopic", "dataTopic")
    CUSTOMBROKER_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    BROKERURL_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    PASSWORD_FIELD_NUMBER: _ClassVar[int]
    UPDATETOPIC_FIELD_NUMBER: _ClassVar[int]
    STATUSTOPIC_FIELD_NUMBER: _ClassVar[int]
    DATATOPIC_FIELD_NUMBER: _ClassVar[int]
    customBroker: bool
    port: int
    brokerUrl: str
    username: str
    password: str
    updateTopic: str
    statusTopic: str
    dataTopic: str
    def __init__(self, customBroker: bool = ..., port: _Optional[int] = ..., brokerUrl: _Optional[str] = ..., username: _Optional[str] = ..., password: _Optional[str] = ..., updateTopic: _Optional[str] = ..., statusTopic: _Optional[str] = ..., dataTopic: _Optional[str] = ...) -> None: ...
